﻿using System;
using System.Linq;
using System.Reflection;
using System.ComponentModel;
using System.Windows.Forms;

namespace ComPortSniffer
{
   /// <summary>
   /// Steuerelement zur Auswahl eines Enumerationswerts.
   /// </summary>
   /// <typeparam name="T">Der Enumrationstyp aus dem ein Wert ausgewählt werden soll.</typeparam>
   /// <remarks>Die Anzeigetexte werden aus dem <see cref="DescriptionAttribute"/> der Enumerationswerte gewonnen.</remarks>
   public partial class EnumSelector<T> : System.Windows.Forms.UserControl where T : struct, IConvertible
   {
      /// <summary>
      /// Klasse zur Darstellung von Items in listenartigen Steuerelementen mit Text und Wert.
      /// </summary>
      protected class ListItem
      {
         /// <summary>Anzuzeigender Text.</summary>
         public readonly string Text;
         /// <summary>Dem Item zu Grunde liegender Wert.</summary>
         public readonly object Tag;

         /// <summary>
         /// Initialisiert eine neue Instanz der <see cref="ListItem"/>-Klasse.
         /// </summary>
         /// <param name="Text">Anzuzeigender Text.</param>
         /// <param name="Tag">Dem Item zu Grunde liegender Wert.</param>
         public ListItem(string Text, object Tag)
         {
            this.Text = Text;
            this.Tag = Tag;
         }

         /// <summary>
         /// Liefert den im listenartigen Steuerelementen anzuzeigenden Text.
         /// </summary>
         /// <returns>Der im listenartigen Steuerelementen anzuzeigende Text.</returns>
         public override string ToString()
         {
            return Text;
         }

         /// <summary>
         /// Konvertiert ein <see cref="ListItem"/>-Objekt in ein Enumerations-Objekt.
         /// </summary>
         /// <typeparam name="T1">Der Enumerations-Typ in den das ListItem konvertiert werden soll.</typeparam>
         /// <param name="l">Ein <see cref="ListItem"/>-Objekt. Typischerweise das SelectedItem-Objekt eines listenartigen Steuerelements.</param>
         /// <returns>Der konvertierte Enumerations-Wert.</returns>
         public static T1 toEnum<T1>(object l)
         {
            ListItem item = (ListItem)l;
            return (T1)item.Tag;
         }

         /// <summary>
         /// Liest das <see cref="DescriptionAttribute"/> des Enumerationswertes aus.
         /// </summary>
         /// <param name="value">Der Enumerationswerte, dessen <see cref="DescriptionAttribute"/> ermittelt werden soll.</param>
         /// <returns>Der beim Enumerationswert hintelegte Beschreibungstext oder der Name des Enumerationswerts.</returns>
         public static string GetEnumDescription(Enum value)
         {
            FieldInfo fi = value.GetType().GetField(value.ToString());

            DescriptionAttribute[] attributes = (DescriptionAttribute[])fi.GetCustomAttributes(typeof(DescriptionAttribute), false);

            if ((!(attributes == null) & attributes.Any()))
               return attributes.First().Description;

            return value.ToString();
         }
      }

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="EnumSelector(Of T)"/>-Klasse.
      /// </summary>
      public EnumSelector()
      {
         if ((!typeof(T).IsEnum))
            throw new ArgumentException("T must be an enumerated type");

         // Dieser Aufruf ist für den Designer erforderlich.
         InitializeComponent();

         foreach (var enumValue in Enum.GetValues(typeof(T)))
            ComboBox1.Items.Add(new ListItem(ListItem.GetEnumDescription((Enum)enumValue), enumValue));
      }

      /// <summary>
      /// Tritt ein, wenn die Selektion geändert wurde.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="enumValue">Der ausgewählte Enumerationswert.</param>
      public event SelectionChangedEventHandler SelectionChanged;

      public delegate void SelectionChangedEventHandler(object sender, T enumValue);


      /// <summary>
      /// Ruft den Index des derzeit ausgewählten Elements ab oder legt diesen fest.
      /// </summary>
      /// <returns>Ein nullbasierter Index des derzeit ausgewählten Elements. Wenn kein Element ausgewählt ist, wird ein Wert von -1 zurückgegeben..</returns>
      public int SelectedIndex
      {
         get
         {
            return ComboBox1.SelectedIndex;
         }
         set
         {
            ComboBox1.SelectedIndex = value;
         }
      }

      /// <summary>
      /// Ruft das derzeit ausgewählte Element ab oder legt es fest.
      /// </summary>
      /// <returns>Das derzeit ausgewählte Element.</returns>
      [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
      public T SelectItem
      {
         get
         {
            return ListItem.toEnum<T>(ComboBox1.SelectedItem);
         }
         set
         {
            for (var i = 0; i <= ComboBox1.Items.Count - 1; i++)
            {
               ListItem li = (ListItem)ComboBox1.Items[i];
               if (ListItem.toEnum<T>(li).Equals(value))
               {
                  ComboBox1.SelectedIndex = i;
                  return;
               }
            }
            throw new InvalidEnumArgumentException();
         }
      }


      /// <summary>Leitet das SelectedIndexChanged-Ereignis weiter.</summary>
      private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
      {
         SelectionChanged?.Invoke(this, ListItem.toEnum<T>(ComboBox1.SelectedItem));
      }

      private void EnumSelector_Load(object sender, EventArgs e)
      {
         // Die Container-Hierrarchie bis zum Form aufsteigen.
         var p = this.Parent;
         FieldInfo toolTipFieldInfo = null;

         while (!p.GetType().IsSubclassOf(typeof(Form)))
            p = p.Parent;

         var bindingFlags = BindingFlags.Instance |
                   BindingFlags.NonPublic |
                   BindingFlags.Public;

         foreach (FieldInfo f in p.GetType().GetFields(bindingFlags))
         {
            if (f.FieldType == typeof(ToolTip))
            {
               toolTipFieldInfo = f;
               break;
            }
         }

         if (toolTipFieldInfo != null)
         {
            var valueOfField = toolTipFieldInfo.GetValue(p);
            var customControlType = toolTipFieldInfo.FieldType;
            var methodInfo = customControlType.GetMethod("GetToolTip");
            var toolTipText =(String) methodInfo.Invoke(valueOfField, new object[] { this });
            toolTip1.SetToolTip(ComboBox1, toolTipText);
         }
      }
   }
}